#include "Arduino.h"
#include "RGB_sprite.h"
#include "masterScreen.h"



/*
| - Description:
|   - Initialize sprite dimensions and what value the sprite holds
|   - ... syntax as the last argument is used along with va_list for
|   - variable arguments
|
| - Parameter:
|   - width   the width of the sprite
|   - height  the height of the sprite
|   - ...     variable length of arguments can follow,
|             example.Sprite(2,4, 1,1,
|                                 2,2,
|                                 3,3,
|                                 4,4);
*/
void RGB_Sprite::Sprite(byte width, byte height, ...)
{
  // dynamically allocating the size of the sprite to save space.
  // Should only be created within the setup()
  init(width, height);
  visible = true;
  //Next x and Next y set to 0, for initialization
  x_ = 0;
  y_ = 0;
  //out of range indicator, means it hasnt been set yet
  origin[0] = 0x80;
  origin[1] = 0x80;

  va_list ap;
  va_start(ap, (width * height));

  for(int row = 0; row < _height; row++){
    for(int col = 0; col < _width; col++){
       //This is where the ... parameters are placed into the sprite matrix.
      _sprite[col][row] = (unsigned char)va_arg(ap, int);
    }
  }
  va_end(ap);
}


/*
| - Description
|   - Get the width of this Sprite
|
| - return The width of this Sprite
*/
byte RGB_Sprite::width() const
{
  return _width;
}

/*
| - Description
|   - Get the height of this Sprite
|
| - return The height of this Sprite
*/
byte RGB_Sprite::height() const
{
  return _height;
}

/*
| - Description
|   - write the value given to the (x,y) coordinates
|   - of the sprite matrix
|
|   -No boundary error checking currently
|
*/
void RGB_Sprite::write(byte x, byte y, byte value)
{
  //if (!_sprite) return;
  //if (x < 0 || x >= _width || y < 0 || y >= _height) return;

  // we need to bitwise-or the value of the other pixels in the byte with
  // the new value, masked and shifted into the proper bits.
  //_sprite[y] = (_sprite[y] & ~(0x01 << x)) | ((value & 0x01) << x);


  _sprite[x][y] = value;

}


/*
| - Description
|   - Get the value at a specific coordinate
|
| - Parameter
|   - x   the x part of the coordinate
|   - y   the y part of the coordinate
*/
byte RGB_Sprite::read(byte x, byte y) const
{
 //if (!_sprite) return 0;
 // if (x < 0 || x >= _width || y < 0 || y >= _height) return 0;

  //return (_sprite[y] >> x) & 0x01;
}


/*
| - Description
|   - Get the value at a specific coordinate
|
| - Parameter
|   - x   the x part of the coordinate
|   - y   the y part of the coordinate
*/
byte RGB_Sprite::getSpriteValue(byte col, byte row)
{/*
  if(row > _height || col > _width){
    return 127;
  }else{
    return _sprite[col][row];
  }
*/

  return _sprite[col][row];
}

/*
| - Description
|   - Print the sprite array to the serial monitor
|   - used for testing
|
*/
void RGB_Sprite::printSprite()
{
  for(int row = 0; row < _height; row++){
    for(int col = 0; col < _width; col++){
      Serial.print(_sprite[col][row]);
    }
    Serial.println("");
  }
}

/*
| - Description
|   - update origin of the sprite, usually performed
|   - after collision detection is passed and before updating
|   - the screen matrix
|
| - Parameter
|   - x   the x part of the coordinate
|   - y   the y part of the coordinate
*/
void RGB_Sprite::updateOrigin(byte x, byte y)
{
  origin[0] = x;
  origin[1] = y;
}


/*
| - Description
|   - takes a masterScreen and tests the new origin of the sprite
|   - againt other sprite dots already on the matrix
|   - less efficent
|
| - Parameter
|   - screen_copy   gameScreen object, allowing access to the
|                     _masterScreen[8][8] array
|   - x             the x part of the coordinate
|   - y             the y part of the coordinate
*/
bool RGB_Sprite::isCollision(masterScreen screen_copy, byte new_x, byte new_y)
{// new_x and new_y cant be less than 0 or greater than 7

  bool result = false;

  //*********Clear self on screen_copy********* clear origin
  for(byte col = origin[1]; col < (origin[1] + _height); col++)
  {
    for(byte row = origin[0]; row < (origin[0] + _width); row++)
    {
      if(row <= 7 && col <= 7){
        screen_copy.modifyScreen(row,col, 0);
      }
    }
  }

  //**********Check New Origin**********
  byte spriteRow = 0;
  byte spriteCol = 0;
  for(byte col = new_y; col < (new_y + _height) ; col++){
    for(byte row = new_x; row < (new_x + _width); row++){
      if((col <= 7) && (row <= 7))
      {
        if(_sprite[spriteRow][spriteCol] != 0){
            if(screen_copy.getScreenValue(row,col) != 0){
              //isCollision? True
              result = true;
              //Serial.println("Taken");
            }
        }
        spriteRow++;
      }
    }
    spriteRow = 0;
    spriteCol++;
  }

  return result;
}




/*
| - Description
|   - tests the new origin of the sprite against the other sprites in a
|     sprite list
|   - first overlapping sprites are checked, then the overlapped sections
|     are tested
|   - against each other to test for a collision
|   - sprites dots with the value of zero will not be considered a
|     collision
|
| - Parameter
|   - num_sprites   number of sprite elements in the sprite list
|   - cur_index     the current index of this sprite in reference to the
|                     sprite list
|   - Sprite_array  list of sprite objects
|   - new_x         the new x part of the coordinate to be tested before
|                     moving sprite
|   - new_y         the new y part of the coordinate to be tested before
|                     moving sprite
|
| - Return
|   - result is the return value, isCollision true means there was a
|       collision
|   - isCollision false, means no collision
*/
bool RGB_Sprite::isCollisionNoScreen(byte num_sprites, byte cur_index, RGB_Sprite Sprite_array[], byte new_x, byte new_y)
{
  bool result = false;
  byte spriteRow = 0;
  byte spriteCol = 0;
  byte new_x_high = (new_x + (Sprite_array[cur_index].width() - 1));
  byte new_y_high = (new_y + (Sprite_array[cur_index].height() - 1));
  byte other_x_low;
  byte other_y_low;
  byte other_x_high;
  byte other_y_high;

  //Serial.println("before for loop ");
  for(byte i = 0; i < num_sprites; i++)
  {
    //Serial.print("start for loop ");
    //Serial.println(i);
    if(i != cur_index){  //skip this test, its the current sprite obj
      if((Sprite_array[i].visible == true) && (visible == true)){
        //Serial.println("visible ");
        other_x_low = Sprite_array[i].origin[0];
        other_y_low = Sprite_array[i].origin[1];
        other_x_high = Sprite_array[i].origin[0]+ (Sprite_array[i].width() - 1);
        other_y_high = Sprite_array[i].origin[1]+ (Sprite_array[i].height() -1);

        if(((new_x <= other_x_high) && (new_x >= other_x_low)) || ((other_x_low <= new_x_high) && (other_x_low >= new_x))){
          //possible collision
          //Serial.println("possible collision ");
          if((new_y <= other_y_high) && (new_y >= other_y_low) || ((other_y_low <= new_y_high) && (other_y_low >= new_y))){
            //sprite overlap, check dots for collision
            //Serial.println("sprite overlap ");
            for(byte row = 0; row <= (other_y_high - new_y); row++){
              //Serial.print("ROW ");
              //Serial.println(row);
              for(byte col = 0; col <= (other_x_high - new_x); col++){
                //Serial.print("COL ");
                //Serial.println(col);
              //(new_x_origin + col) - other_x_origin)
                if((_sprite[col][row] != 0) && (Sprite_array[i].getSpriteValue(((new_x-col)-other_x_low), ((new_y-row)-other_y_low)) != 0)){
                //this should be a collision.
                  //Serial.println("COLLISION ");
                  result = true;
                }
              }
            }
          }
        }
      }
    }
  }
  return result;
}

/*
| - Description
|   - is sprite colliding with the Left side of the board?
|
| - Return
|   - bool   True, collision with Left boundry
|            False, No collision with left boundry
|
*/
bool RGB_Sprite::isL_boardCollision()
{
  if(origin[0] <= 0){
    return true;
  }
  return false;
}

/*
| - Description
|   - is sprite colliding with the Right side of the board?
|
| - Return
|   - bool   True, collision with Right boundry
|            False, No collision with Right boundry
|
*/
bool RGB_Sprite::isR_boardCollision()
{
  //Right edge of sprite collide with right boundry
  if((origin[0] + (_width - 1)) >= 7){
    return true;
  }
  return false;
}

/*
| - Description
|   - is sprite colliding with the Top side of the board?
|
| - Return
|   - bool   True, collision with Top boundry
|            False, No collision with Top boundry
|
*/
bool RGB_Sprite::isT_boardCollision()
{
  //top edge of sprite collides with boundry
  if((origin[1] + (_width - 1)) >= 7){
    return true;
  }
  return false;
}

/*
| - Description
|   - is sprite colliding with the Bottom side of the board?
|
| - Return
|   - bool   True, collision with Bottom boundry
|            False, No collision with Bottom boundry
|
*/
bool RGB_Sprite::isB_boardCollision()
{
  if(origin[1] <= 0){
    return true;
  }
  return false;
}




///**************** Private Methods **********************

/*
| - Description
|   - Dynamically Allocate the sprite to the given dimensions of sprite
|   - maximum size of 8x8. The sprite is dynamically created, but not
|       distroyed.
|   - no resizing is needed
|
*/
void RGB_Sprite::init(byte width, byte height)
{
  // limiting sprites to 8x8 the size of the display
  _width = width >= 8 ? 8 : width;    //Row
  _height = height >= 8 ? 8 : height; //Col

  //Allocation
  _sprite = new byte*[_width];
  for(byte col = 0; col < _width; col++){
    _sprite[col] = new byte[_height];
  }

  // Initialization
  for(byte row = 0; row < _height; row++){
    for(byte col = 0; col < _width; col++){
      _sprite[col][row] = 0;
    }
  }
}
